# Python wrapper for libsoxr
# https://github.com/dofuuz/python-soxr

import numpy as np
cimport numpy as np

cimport csoxr


cdef class CySoxr:
    cdef csoxr.soxr_t _soxr
    cdef double _in_rate
    cdef double _out_rate
    cdef type _dtype

    def __cinit__(self, double in_rate, double out_rate, unsigned num_channels, type dtype):
        self._in_rate = in_rate
        self._out_rate = out_rate
        self._dtype = dtype

        cdef csoxr.soxr_datatype_t io_type
        if dtype == np.float32:
            io_type = csoxr.SOXR_FLOAT32_I
        elif dtype == np.float64:
            io_type = csoxr.SOXR_FLOAT64_I
        elif dtype == np.int32:
            io_type = csoxr.SOXR_INT32_I
        elif dtype == np.int16:
            io_type = csoxr.SOXR_INT16_I
        else:
            raise ValueError('Dtype not support')

        cdef csoxr.soxr_io_spec_t io_spec = csoxr.soxr_io_spec(io_type, io_type)

        self._soxr = csoxr.soxr_create(in_rate, out_rate, num_channels, NULL, &io_spec, NULL, NULL)
        if self._soxr is NULL:
            raise MemoryError()

    def __dealloc__(self):
        csoxr.soxr_delete(self._soxr)

    cpdef np.ndarray process(self, np.ndarray x, bint eof=False):
        if 2 < x.ndim:
            raise ValueError('Input must be 1-D or 2-D array')

        cdef size_t ilen = x.shape[0]
        cdef size_t olen = np.ceil(ilen * self._out_rate / self._in_rate)
        cdef unsigned channels = 1
        if 2 == x.ndim:
            channels = x.shape[1]

        dtype = x.dtype
        if dtype != self._dtype:
            raise ValueError('Dtype not match')

        x = np.ascontiguousarray(x)
        cdef np.ndarray out_buf = np.zeros([olen, channels], dtype=dtype, order='c')
        cdef size_t odone

        csoxr.soxr_process(
            self._soxr,
            x.data, ilen, NULL,
            out_buf.data, olen, &odone)

        out_buf = out_buf[:odone]

        cdef np.ndarray eof_buf
        cdef int delay
        if (eof):
            delay = int(csoxr.soxr_delay(self._soxr) + .5)
            eof_buf = np.zeros([delay, channels], dtype=dtype, order='c')
            csoxr.soxr_process(
                self._soxr,
                NULL, 0, NULL,
                eof_buf.data, delay, &odone)

            eof_buf = eof_buf[:odone]

            out_buf = np.vstack([out_buf, eof_buf])

        return out_buf

        # return np.copy(self._out_buf[:odone])


cpdef np.ndarray cysoxr_oneshot(double in_rate, double out_rate, np.ndarray x):
    if 2 < x.ndim:
        raise ValueError('Input must be 1-D or 2-D array')

    cdef size_t ilen = x.shape[0]
    cdef size_t olen = np.ceil(ilen * out_rate / in_rate)
    cdef unsigned channels = 1
    if 2 == x.ndim:
        channels = x.shape[1]

    dtype = x.dtype

    cdef csoxr.soxr_datatype_t io_type
    if dtype == np.float32:
        io_type = csoxr.SOXR_FLOAT32_I
    elif dtype == np.float64:
        io_type = csoxr.SOXR_FLOAT64_I
    elif dtype == np.int32:
        io_type = csoxr.SOXR_INT32_I
    elif dtype == np.int16:
        io_type = csoxr.SOXR_INT16_I
    else:
        raise ValueError('Dtype not support')

    cdef csoxr.soxr_io_spec_t io_spec = csoxr.soxr_io_spec(io_type, io_type)
    cdef csoxr.soxr_quality_spec_t quality = csoxr.soxr_quality_spec(csoxr.SOXR_HQ, 0)

    x = np.ascontiguousarray(x)    # make array C-contiguous

    cdef size_t odone
    cdef np.ndarray out_ndarray = np.zeros([olen, channels], dtype=dtype, order='c')

    csoxr.soxr_oneshot(
        in_rate, out_rate, channels,
        x.data, ilen, NULL,
        out_ndarray.data, olen, &odone,
        &io_spec, &quality, NULL)

    return out_ndarray[:odone]



# Python wrapper for libsoxr
# https://github.com/dofuuz/python-soxr


from .cysoxr import CySoxr, cysoxr_oneshot
import numpy as np


class ResampleStream():
    def __init__(self, in_rate: float, out_rate: float, num_channels: int, dtype=np.float32):
        # internally uses Sclar types, not dtype
        if type(dtype) != type:
            dtype = np.dtype(dtype).type
        if not dtype in (np.float32, np.float64, np.int16, np.int32):
            raise ValueError("dtype must be one of ['float32', 'float64', 'int16', 'int32'] and not {}".format(dtype))

        self._type = dtype

        self._cysoxr = CySoxr(in_rate, out_rate, num_channels, self._type)

    def resample_chunk(self, x, end=False):
        if type(x) != np.ndarray or x.dtype.type != self._type:
            x = np.asarray(x, dtype=self._type)

        return self._cysoxr.process(x, end)


def resample(x, in_rate: float, out_rate: float):
    if type(x) != np.ndarray:
        x = np.asarray(x, dtype=np.float32)

    if not x.dtype.type in (np.float32, np.float64, np.int16, np.int32):
        raise ValueError("dtype must be one of ['float32', 'float64', 'int16', 'int32'] and not {}".format(x.dtype.type))

    return cysoxr_oneshot(in_rate, out_rate, x)



# distutils: language=C

# Cython wrapper for libsoxr
# https://github.com/dofuuz/python-soxr

cdef extern from 'soxr.h':
    ctypedef struct soxr_io_spec_t:
        pass
    ctypedef struct soxr_quality_spec_t:
        pass
    ctypedef struct soxr_runtime_spec_t:
        pass

    # --------------------------- Type declarations ----------------------------

    ctypedef struct soxr:
        pass
    ctypedef soxr * soxr_t
    ctypedef const char * soxr_error_t                 # 0:no-error; non-0:error.

    ctypedef const void * soxr_in_t
    ctypedef void       * soxr_out_t

    # --------------------------- API main functions ---------------------------

    cdef const char * soxr_version()   # Query library version: "libsoxr-x.y.z"

    # Create a stream resampler:
    cdef soxr_t soxr_create(
        double      input_rate,      # Input sample-rate.
        double      output_rate,     # Output sample-rate.
        unsigned    num_channels,    # Number of channels to be used.
            # All following arguments are optional (may be set to NULL).
        soxr_error_t *,              # To report any error during creation.
        const soxr_io_spec_t *,      # To specify non-default I/O formats.
        const soxr_quality_spec_t *, # To specify non-default resampling quality.
        const soxr_runtime_spec_t *) # To specify non-default runtime resources.
    # Default io_spec      is per soxr_io_spec(SOXR_FLOAT32_I, SOXR_FLOAT32_I)
    # Default quality_spec is per soxr_quality_spec(SOXR_HQ, 0)
    # Default runtime_spec is per soxr_runtime_spec(1)                          */

    # If not using an app-supplied input function, after creating a stream
    # resampler, repeatedly call:
    cdef soxr_error_t soxr_process(
        soxr_t      resampler,      # As returned by soxr_create.
                                # Input (to be resampled):
        soxr_in_t   in_,            # Input buffer(s); may be NULL (see below).
        size_t      ilen,           # Input buf. length (samples per channel).
        size_t      * idone,        # To return actual # samples used (<= ilen).
                                # Output (resampled):
        soxr_out_t  out,            # Output buffer(s).
        size_t      olen,           # Output buf. length (samples per channel).
        size_t      * odone)        # To return actual # samples out (<= olen).
    # Note that no special meaning is associated with ilen or olen equal to
    # zero.  End-of-input (i.e. no data is available nor shall be available)
    # may be indicated by seting `in' to NULL.

    # Common stream resampler operations:
    cdef soxr_error_t soxr_error(soxr_t)   # Query error status.
    cdef size_t   * soxr_num_clips(soxr_t) # Query int. clip counter (for R/W).
    cdef double     soxr_delay(soxr_t)  # Query current delay in output samples.
    cdef const char * soxr_engine(soxr_t)  # Query resampling engine name.

    cdef soxr_error_t soxr_clear(soxr_t) # Ready for fresh signal, same config.
    cdef void         soxr_delete(soxr_t)  # Free resources.

    # `Short-cut', single call to resample a (probably short) signal held entirely
    # in memory.  See soxr_create and soxr_process above for parameter details.
    # Note that unlike soxr_create however, the default quality spec. for
    # soxr_oneshot is per soxr_quality_spec(SOXR_LQ, 0).
    cdef soxr_error_t soxr_oneshot(
        double         input_rate,
        double         output_rate,
        unsigned       num_channels,
        soxr_in_t    in_, size_t ilen, size_t * idone,
        soxr_out_t   out, size_t olen, size_t * odone,
        const soxr_io_spec_t *,
        const soxr_quality_spec_t *,
        const soxr_runtime_spec_t *)

    cdef enum soxr_datatype_t:          # Datatypes supported for I/O to/from the resampler: */
        # Internal; do not use:
        SOXR_FLOAT32, SOXR_FLOAT64, SOXR_INT32, SOXR_INT16, SOXR_SPLIT = 4,

        # Use for interleaved channels:
        SOXR_FLOAT32_I = SOXR_FLOAT32, SOXR_FLOAT64_I, SOXR_INT32_I, SOXR_INT16_I,

        # Use for split channels:
        SOXR_FLOAT32_S = SOXR_SPLIT  , SOXR_FLOAT64_S, SOXR_INT32_S, SOXR_INT16_S

    # -------------------------- API type constructors -------------------------

    cdef soxr_quality_spec_t soxr_quality_spec(
        unsigned long recipe,       # Per the #defines immediately below.
        unsigned long flags)        # As soxr_quality_spec_t.flags.

    cdef enum soxr_quality_recipe_t:
        SOXR_QQ
        SOXR_LQ
        SOXR_MQ
        SOXR_HQ
        SOXR_VHQ

    cdef soxr_io_spec_t soxr_io_spec(
        soxr_datatype_t itype,
        soxr_datatype_t otype)



from setuptools import find_packages
from setuptools import setup, Extension


with open("README.md", "r") as fh:
    long_description = fh.read()


class CySoxrExtension(Extension):
    def __init__(self, *args, **kwargs):
        self._include = []
        super().__init__(*args, **kwargs)

    @property
    def include_dirs(self):
        import numpy
        return self._include + [numpy.get_include()]

    @include_dirs.setter
    def include_dirs(self, dirs):
        self._include = dirs

src = [
    'libsoxr/src/soxr.c',
    'libsoxr/src/data-io.c',
    'libsoxr/src/dbesi0.c',
    'libsoxr/src/filter.c',
    'libsoxr/src/fft4g64.c',
    'libsoxr/src/cr.c',

    # WITH_CR32
    'libsoxr/src/cr32.c',
    'libsoxr/src/fft4g32.c',

    # WITH_CR64
    'libsoxr/src/cr64.c',

    # WITH_VR32
    # 'libsoxr/src/vr32.c',

    # WITH_CR32S
    'libsoxr/src/cr32s.c',
    'libsoxr/src/pffft32s.c',
    'libsoxr/src/util32s.c',

    # WITH_CR64S
    # 'libsoxr/src/cr64s.c',
    # 'libsoxr/src/pffft64s.c',
    # 'libsoxr/src/util64s.c',

    # Cython wrapper
    'csoxr.pxd',
    'cysoxr.pyx'
]

extensions = [
    CySoxrExtension(
        "cysoxr",
        src,
        include_dirs=['libsoxr/src', 'soxr'],
        language="c",
        extra_compile_args=['-DSOXR_LIB'])
]

setup(
    name="soxr",
    version="0.0.3",
    author="dofuuz",
    description="High quality, one-dimensional sample-rate conversion library",
    keywords='samplerate, SRC',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dofuuz/python-soxr",
    packages=find_packages(),
    ext_package='soxr',
    ext_modules=extensions,
    python_requires='>=3.5',
    install_requires=['numpy'],
    tests_require=['pytest'],
    classifiers=[
        "Development Status :: 3 - Alpha",

        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Telecommunications Industry",
        "Topic :: Multimedia :: Sound/Audio :: Analysis",
        "Topic :: Multimedia :: Sound/Audio :: Conversion",
        "Topic :: Scientific/Engineering",

        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",

        "Programming Language :: C",
        "Programming Language :: Python",
    ],
    license_files=['LICENSE.txt'],
)

